
assert( LibStub, "LibStub is not loaded." )

local addon = LibStub( "AceAddon-3.0" ):NewAddon( "hcGossipBot", "AceConsole-3.0", "AceEvent-3.0" )

local DEFAULTS = {
	profile = {
		-- True if this addon should be enabled.
		enabled = true,
		-- True if debugging message should be output by this addon.
		verbose = false, 

		-- When true, gossip interactions with NPC's will be automated.
		gossip = true,
		-- When true, gossip text will be output to the default chat frame 
		-- as if the NPC was speaking.
		gossipTalk = true,

		-- When true, available quests will be automatically accepted.
		questAccept = true,
		-- When true, active quests will be automatically completed.
		questComplete = true,
		-- When "price", the most valuable quest reward will be 
		-- automatically chosen. When nil, no quest reward will be 
		-- automatically chosen.
		questReward = nil,
		
		-- When true, newly-accepted quests will automatically be shared 
		-- while in a party.
		shareParty = false,
		-- When true, newly-accepted quests will automatically be shared 
		-- while in a raid.
		shareRaid = false,
		
	},
}

local function _IsAddonDisabled( info )
	return not addon.db.profile.enabled
end

local OPTIONS = {
	type = "group",
	name = "Gossip",
	args = {
		enable = {
			type = "toggle",
			order = 1,
			width = "full",
			name = "Enabled",
			desc = "Enables or disables automatic NPC interaction.",
			descStyle = "inline",
			get = function( info ) return addon.db.profile.enabled end,
			set = function( info, value ) if value then addon:Enable( ) else addon:Disable( ) end end,
			disabled = false,
		},
		verbose = {
			type = "toggle",
			order = 2,
			width = "full",
			name = "Verbose",
			desc = "Enables or disables verbose output messages.",
			descStyle = "inline",
			get = function( info ) return addon.db.profile.verbose end,
			set = function( info, value ) addon.db.profile.verbose = value end,
			disabled = _IsAddonDisabled,
		},
	},
}

local function _SetDefaultOptions( )
	addon.db:ResetProfile( )
	LibStub( "AceConfigRegistry-3.0" ):NotifyChange( addon.name )
end

local function CreateProfileOptionFrame( addon )
	addon.profiles = LibStub( "AceDBOptions-3.0" ):GetOptionsTable( addon.db )
	LibStub( "AceConfig-3.0" ):RegisterOptionsTable( addon.name .. "Profiles", addon.profiles )
	LibStub( "AceConfigDialog-3.0" ):AddToBlizOptions( addon.name .. "Profiles", "Profiles", addon.name )
end

function addon:OnInitialize( )
	self.db = LibStub( "AceDB-3.0" ):New( "hcGossipBotSavedVariables", DEFAULTS, true )
	
	LibStub( "AceConfig-3.0" ):RegisterOptionsTable( self.name, OPTIONS )
	local optionsFrame = LibStub( "AceConfigDialog-3.0" ):AddToBlizOptions( self.name, self.name )
	optionsFrame.default = _SetDefaultOptions
 
	-- CreateAboutOptionFrame( self.name )
	CreateProfileOptionFrame( self )

	-- TODO document me
	self.handler = {
		disabled = false,
	}

	self:SetEnabledState( not _IsAddonDisabled( ) )
end

local EVENTS = {
	"GOSSIP_CLOSED",
	"GOSSIP_SHOW",
	"QUEST_ACCEPTED",
	"QUEST_ACCEPT_CONFIRM",
	"QUEST_AUTOCOMPLETE",
	"QUEST_COMPLETE",
	"QUEST_DETAIL",
	"QUEST_PROGRESS",
}

function addon:OnEnable( )
	self.db.profile.enabled = true
	for _, event in ipairs( EVENTS ) do
		self:RegisterEvent( event )
	end
end

function addon:OnDisable( )
	self.db.profile.enabled = false
end

--- Tests whether or not interaction with the current NPC is disabled.
-- If a modifier key (CTRL, SHIFT, or ALT) is pressed, then automation 
-- is suspended for any remaining interaction with the active NPC.
-- @return true if the interaction handler should be active.
function addon:IsHandlerDisabled( )
	self.handler.disabled = IsModifierKeyDown( ) or self.handler.disabled
	return self.handler.disabled or IsModifierKeyDown( )
end

--- Resets the interaction to being enabled.
function addon:ResetHandler( )
	self.handler.disabled = false
end

--- This event fires when interaction with an NPC ends.
-- This is not fired for some questgiver NPC's.
function addon:GOSSIP_CLOSED( event )
	self:ResetHandler( )
end

local function _ShouldCompleteQuest( isComplete )

	if not addon.db.profile.questComplete then
		-- TODO output something?
		return false
	end

	if not isComplete then
		-- TODO output something?
		return false
	end
	
	return true
end

local function _FindCompleteQuest( ... )
	local index = 1
	for i = 1, select( "#", ... ), 4 do
		local isComplete = select( i + 3, ... )
		if _ShouldCompleteQuest( isComplete ) then
			return index
		end
		index = index + 1
	end
	
	return 0
end

--- Returns true if the player should accept a quest.
-- This returns false if automatic accepting of quests is disabled, if
-- the quest log is full, or if the maximum number of daily quests has 
-- already been completed while checking a daily quest.
-- @param true if the quest being checked is a daily quest.
local function _ShouldAcceptQuest( isDaily )
	if not addon.db.profile.questAccept then
		return false
	end
	
	local _, numQuests = GetNumQuestLogEntries( )
	if numQuests >= MAX_QUESTS then
		return false
	end
	
	if isDaily and ( GetDailyQuestsCompleted( ) >= GetMaxDailyQuests( ) ) then
		return false
	end
	
	return true
end

local function _FindAcceptQuest( ... )
	local index = 1
	for i = 1, select( "#", ... ), 5 do
		local isDaily = select( i + 3, ... )
		if _ShouldAcceptQuest( isDaily ) then
			return index
		end
		index = index + 1
	end
	
	return 0
end

--- Fires when an NPC gossip interaction begins.
-- This is fired for most initial NPC interactions. 
function addon:GOSSIP_SHOW( event )
	if self:IsHandlerDisabled( ) then
		return
	end
	
	-- If any active quests have been completed, turn in the first one 
	-- the active quest list.
	if self.db.profile.questComplete then
		if GetNumGossipActiveQuests( ) > 0 then
			local index = _FindCompleteQuest( GetGossipActiveQuests( ) )
			if index > 0 then
				SelectGossipActiveQuest( index )
				return
			end
		end
	end
	
	-- If any quests are available, accept the first one in the available
	-- quest list.
	if self.db.profile.questAccept then
		if GetNumGossipAvailableQuests( ) > 0 then
			local index = _FindAcceptQuest( GetGossipAvailableQuests( ) )
			if index > 0 then
				SelectGossipAvailableQuest( index )
				return
			end
		end
	end

	-- If only one gossip option is available, select it.
	if GetNumGossipOptions( ) == 1 and self.db.profile.gossip then
		self:PrintG( true, ( GetGossipOptions( ) ) )
		self:PrintG( false, GetGossipText( ) )
		SelectGossipOption( 1 )
		return
	end
	
	if ( GetNumGossipOptions( ) + GetNumGossipAvailableQuests( ) + GetNumActiveQuests( ) == 0 ) and self.db.profile.gossip then
		-- This seems to fire for merchants
		if not ( GetMerchantNumItems( ) > 0 or CanMerchantRepair( ) ) then
			self:PrintG( false, GetGossipText( ) )
		end
		CloseGossip( )
		return
	end
end

--- Tests whether or not quests should be shared.
-- Returns true when the player is in a raid and sharing with a raid 
-- is enabled, or when the player is in a party and sharing with a 
-- party is enabled. 
-- @return true if quests should be shared.
local function _ShouldShareQuest( )
	if GetNumRaidMembers( ) > 0 and addon.db.profile.shareRaid then
		return true
	end
	
	if GetNumPartyMembers( ) > 0 and addon.db.profile.shareParty then
		return true
	end
	
	return false
end

---	Fires when a new quest is added to the player's quest log.
-- This happens after a player accepts a quest.
-- @param questIndex index where the accepted quest was placed in the quest log.
function addon:QUEST_ACCEPTED( event, questIndex )
	if self:IsHandlerDisabled( ) then
		return
	end

	if _ShouldShareQuest( self.db.profile ) then
		SelectQuestLogEntry( questIndex )
		if GetQuestLogPushable( ) then
			QuestLogPushQuest( questIndex )
		end
	end
end

---	Fires when escort quests are started by a party member.
-- @param name the name of the user who started the quest.
-- @param quest te name of the quest that was started.
function addon:QUEST_ACCEPT_CONFIRM( event, name, quest )
	if self:IsHandlerDisabled( ) then
		return
	end
	
	if _ShouldAcceptQuest( ) then
		ConfirmAcceptQuest( )
		self:PrintV( "Accepted quest '" .. quest .. "' started by " .. name .. "." )
	end
end

---	Fires when a quest is automatically completed.
-- These types of quests may be handed in remotely, and will pop up 
-- with a completion indicator in the objective list.
-- @param questId the id of the quest completed.
function addon:QUEST_AUTOCOMPLETE( event, questId )
	if self:IsHandlerDisabled( ) then
		return
	end
	
	if _ShouldCompleteQuest( true ) then
		ShowQuestComplete( questId )
	end
end

--- Returns the vendor price of a given item.
-- @param item an item id, item name, or item link
-- @return the vendor price, in copper.
local function _GetItemVendorPrice( item )
	local _, _, _, _, _, _, _, _, _, _, vendorPrice = GetItemInfo( item )
	return vendorPrice
end

--- Returns the quest item choice with the highest vendor value.
-- @return the choice index of the highest vendor value item.
local function _GetBestQuestItemChoice( )
	local bestChoice = nil
	local bestPrice = nil
	for choice = 1, GetNumQuestChoices( ) do
		local item = GetQuestItemLink( "choice", choice )
		local price = _GetItemVendorPrice( item )
		if ( not bestPrice ) or ( bestPrice < price ) then
			bestChoice = choice
			bestPrice = price
		end
	end
	return bestChoice
end

---	Fires when the player is looking at the Complete page for a quest.
-- This may be at a questgiver, or may be via remote completion.
function addon:QUEST_COMPLETE( event )
	if self:IsHandlerDisabled( ) then
		return
	end
	
	if not _ShouldCompleteQuest( true ) then
		return
	end

	local choice = nil
	
	if GetNumQuestChoices( ) > 0 then
		if self.db.profile.questReward == "price" then
			choice = _GetBestQuestItemChoice( )
		else
			return
		end
	end
	
	self:PrintG( false, GetProgressText( ) )
	GetQuestReward( choice )
end

--- Fires when details of an available quest are presented by a questgiver.
function addon:QUEST_DETAIL( event )
	if self:IsHandlerDisabled( ) then
		return
	end
	
	if not _ShouldAcceptQuest( QuestIsDaily( ) ) then
		DeclineQuest( )
		return
	end

	self:PrintG( false, GetQuestText( ), GetObjectiveText( ) )
	AcceptQuest( )
end

--- Fires when interacting with a questgiver about an active quest.
function addon:QUEST_PROGRESS( event )
	if self:IsHandlerDisabled( ) then
		return
	end
	
	if _ShouldCompleteQuest( IsQuestCompletable( ) ) then
			self:PrintG( false, GetProgressText( ) )
		-- TODO output choices?
		CompleteQuest( )
		return
	end
end

local function _GetMessageTypeColorHex( messageType )
	local r, g, b, _ = GetMessageTypeColor( messageType )
	return string.format( "|cff%02x%02x%02x", r * 255, g * 255, b * 255 )
end

local function _GetClassTypeColorHex( classType )
	local color = RAID_CLASS_COLORS[ classType ]
	return string.format( "|cff%02x%02x%02x", color.r * 255, color.g * 255, color.b * 255 )
end

local function _CreateNpcEmotePrefix( )
	local textColor = _GetMessageTypeColorHex( "CHAT_MSG_EMOTE" )
	return textColor
end

local function _CreateEmotePrefix( player )
	return _CreateNpcEmotePrefix( )
end

local function _CreateNpcGossipPrefix( )
	local textColor = _GetMessageTypeColorHex( "CHAT_MSG_MONSTER_SAY" )
	local guid = UnitGUID( "npc" )
	local name = UnitName( "npc" )
	return string.format( "%s|Hunit:%s:%s|h%s|h says: ", textColor, guid, name, name )
end

local function _CreatePlayerGossipPrefix( )
	local textColor, _ = _GetMessageTypeColorHex( "CHAT_MSG_SAY" )
	local name = UnitName( "player" )
	local _, classType = UnitClass( "player" )
	local classColor = _GetClassTypeColorHex( classType )
	return string.format( "%s|Hplayer:%s|h[%s]|h%s says: ", classColor, name, name, textColor )
end

local function _CreateGossipPrefix( player )
	if player then
		return _CreatePlayerGossipPrefix( )
	else
		return _CreateNpcGossipPrefix( )
	end
end

local function _PrintGossipItems( player, frame, ... )
	local emotePrefix = _CreateNpcEmotePrefix( )
	local gossipPrefix = _CreateGossipPrefix( player )
  for i = 1, select( '#', ... ) do
		local message = strtrim( select( i, ... ) )
		if strlen( message ) > 0 then
			local emote = string.match( message, "^<(.+)>$" )
			if emote then
				frame:AddMessage( emotePrefix .. emote )
			else
				frame:AddMessage( gossipPrefix .. message )
			end
		end
	end
end

--- Prints gossip text to the default chat frame.
-- If the gossip text contains multiple lines, each line will be output
-- individually. Each line of text is formatted to appear like normal 
-- creature speech messages.
-- @param player true if the gossip text is spoken by the player.
-- @param ... strings to be written.
function addon:PrintG( player, ... )
	if not self.db.profile.gossipTalk then
		return
	end
	
	for i = 1, select( "#", ... ) do
		local message = select( i, ... )
		if type( message ) == "string" then
			_PrintGossipItems( player, DEFAULT_CHAT_FRAME, strsplit( "\n", message ) )
		end
	end
end

--- Prints verbose messages.
-- These messages will only be printed if verbose output is enabled.
-- @param ... strings to be written.
function addon:PrintV( ... )
	if not self.db.profile.verbose then
		return
	end
	self:Print( ... )
end
